searchNodes=[{"doc":"Exhort is an idomatic Elixir interface to the Google OR Tools . Exhort is currently focused on the &quot;SAT&quot; portion of the tooling: A constraint programming solver that uses SAT (satisfiability) methods. The primary API for Exhort is through a few modules: Exhort.SAT.Builder - The module and struct for building a Exhort.SAT.Model . The builder provides functions for defining variables, expressions and building the model. Exhort.SAT.Expr - A factory for expressions, constraints and variables. This module may be used as the primary interface for defining the parts of a model which are then added to a %Exhort.SAT.Builder{} struct before building the model. Exhort.SAT.Model - The result of building a model through Exhort.SAT.Builder.build/1 . Solving the model is done with Exhort.SAT.Model.solve/1 or Exhort.SAT.Model.solve/2 . The latter accepts a function that receives intermediate results in the solution. Exhort.SAT.SolverResponse - A model solution. The %Exhort.SAT.SolverResponse{} struct containts meta-level information of the solution. The module has functions for retriving the values of variables defined in the model. Livebook See the included sample Livebook notebooks for examples on using Exhort. Setup See the Exhort README for information on using Exhort in a Livebook or adding it as a dependency to a project. Exhort uses native code so the host systsem must have a C/C++ compiler and the make utility. API Exhort is in the early stages of development. As such, we are investigating a varity of API approaches. We may end up with more than one (a la Ecto), but in the short term will likely focus on a single approach. The API is centered around the Builder and Expr modules. Those modules leverage Elixir macros to provide a DSL &quot;expression language&quot; for Exhort. Builder Building a model is done Exhort.SAT.Builder . Exhort.SAT.Builder has functions for defining variables, specifying constraints and creating a %Exhort.SAT.Model{} using the build function. By specifying use Exhort.SAT.Builder , all of the relevant modules will be aliased and the Exhort macros will be expanded. use Exhort.SAT.Builder ... builder = Builder . new ( ) |&gt; Builder . def_int_var ( &quot;x&quot; , { 0 , 10 } ) |&gt; Builder . def_int_var ( &quot;y&quot; , { 0 , 10 } ) |&gt; Builder . def_bool_var ( &quot;b&quot; ) |&gt; Builder . constrain ( &quot;x&quot; &gt;= 5 , if : &quot;b&quot; ) |&gt; Builder . constrain ( &quot;x&quot; &lt; 5 , unless : &quot;b&quot; ) |&gt; Builder . constrain ( &quot;x&quot; + &quot;y&quot; == 10 , if : &quot;b&quot; ) |&gt; Builder . constrain ( &quot;y&quot; == 0 , unless : &quot;b&quot; ) { response , acc } = builder |&gt; Builder . build ( ) |&gt; Model . solve ( fn _response , nil -&gt; 1 _response , acc -&gt; acc + 1 end ) # 2 responses acc |&gt; IO . inspect ( label : &quot;acc: &quot; ) response |&gt; IO . inspect ( label : &quot;response: &quot; ) # :optimal response . status |&gt; IO . inspect ( label : &quot;status: &quot; ) # 10, 0, true SolverResponse . int_val ( response , &quot;x&quot; ) |&gt; IO . inspect ( label : &quot;x: &quot; ) SolverResponse . int_val ( response , &quot;y&quot; ) |&gt; IO . inspect ( label : &quot;y: &quot; ) SolverResponse . bool_val ( response , &quot;b&quot; ) |&gt; IO . inspect ( label : &quot;b: &quot; ) See below for more about the expression language used in Exhort. Expr Sometimes it may be more convenient to build up expressions separately and then add them to a %Builer{} all at once. This is often the case when more complex data sets are invovled in generating many variables and constraints for the model. Instead of having to maintain the builder through an Enum.reduce/3 construct like this: builder = Enum . reduce ( all_days , builder , fn day , builder -&gt; Enum . reduce ( all_shifts , builder , fn shift , builder -&gt; shift_option_vars = shifts |&gt; Enum . filter ( fn { _n , d , s } -&gt; d == day and s == shift end ) |&gt; Enum . map ( fn { n , d , s } -&gt; &quot;shift_ \#{ n } _ \#{ d } _ \#{ s } &quot; end ) Builder . constrain ( builder , sum ( shift_option_vars ) == 1 ) end ) end ) Exhort allows the generation of lists of variables or constraint, maybe using Enum.map/2 : shift_nurses_per_period = Enum . map ( all_days , fn day -&gt; Enum . map ( all_shifts , fn shift -&gt; shift_options = Enum . filter ( shifts , fn { _n , d , s } -&gt; d == day and s == shift end ) shift_option_vars = Enum . map ( shift_options , fn { n , d , s } -&gt; &quot;shift_ \#{ n } _ \#{ d } _ \#{ s } &quot; end ) Expr . new ( sum ( shift_option_vars ) == 1 ) end ) end ) |&gt; List . flatten ( ) These may then be added to the builder as a list: builder |&gt; Builder . add ( shift_nurses_per_period ) ... Variables Model variables in the expression language are symbolic, represented as strings or atoms, and so don't interfere to the surrounding Elixir context. This allows the variables to be consistently referenced through a builder pipeline, for example, without having to capture an intermediate result. Elixir variables may be used &quot;as is&quot; in expressions, allowing variables to be generated from enumerable collections. In the following expression, &quot;x&quot; is a model variable, while y is an Elixir variable: &quot;x&quot; &lt; y + 3 Variables may be defined in a few ways. It's often convenient to just focus on the Exhort.SAT.Expr and Exhort.SAT.Builder modules, which each have functions like def_int_var and def_bool_var . all_bins |&gt; Enum . map ( fn bin -&gt; Expr . def_bool_var ( &quot;slack_ \#{ bin } &quot; ) end ) However, BoolVar.new/1 and IntVar.new/1 may also be used: all_bins |&gt; Enum . map ( fn bin -&gt; BoolVar . new ( &quot;slack_ \#{ bin } &quot; ) end ) Of course, such names are still usable in expressions: Expr . new ( &quot;slack_ \#{ bin } &quot; &lt;= bin_total ) Note that any variables or expressions created outside of the Exhort.SAT.Builder still need to be added to a %Exhort.SAT.Builder{} struct for them to be part of the model resulting from build/1 . There's no magic here, these are still Elixir immutable data structures. variables = ... expressions = ... Builder . new ( ) |&gt; Builder . add ( variables ) |&gt; Builder . add ( expressions ) |&gt; Builder . build ( ) Expressions Exhort supports a limited set of expressions. Expressions may use the binary operators + , - and * , with their traditional mathematical meaning. They may also use comparison operators &lt; , &lt;= , == , &gt;= , &gt; , the sum function and even the for comprehension. all_bins |&gt; Enum . map ( fn bin -&gt; vars = Enum . map ( items , &amp; { elem ( &amp;1 , 0 ) , &quot;x_ \#{ elem ( &amp;1 , 0 ) } _ \#{ bin } &quot; } ) load_bin = &quot;load_ \#{ bin } &quot; Expr . constrain ( sum ( for { item , x } &lt;- vars , do : item * x ) == load_bin ) end ) Model The model is the result of finalizing the builder, created through the Exhort.SAT.Builder.build/1 function. The model may then be solved with Exhort.SAT.Model.solve/1 or Exhort.SAT.Model.solve/2 . The latter function allows for a function to be passed to receive intermediate solutions from the solver. SolverResponse The result of Exhort.SAT.Model.solve/1 is a %Exhort.SAT.SolverResponse{} . The response containts meta-level information of the solution. Exhort.SAT.SolverResponse has functions for retriving the values of variables defined in the model. response = Builder . new ( ) |&gt; Builder . def_int_var ( &quot;r&quot; , { 0 , 100 } ) |&gt; Builder . def_int_var ( &quot;p&quot; , { 0 , 100 } ) |&gt; Builder . constrain ( &quot;r&quot; + &quot;p&quot; == 20 ) |&gt; Builder . constrain ( 4 * &quot;r&quot; + 2 * &quot;p&quot; == 56 ) |&gt; Builder . build ( ) |&gt; Model . solve ( ) assert :optimal = response . status assert 8 == SolverResponse . int_val ( response , &quot;r&quot; ) assert 12 == SolverResponse . int_val ( response , &quot;p&quot; ) Implementation Exhort relies on the underlying native C++ implementation of the Google OR Tools. Exhort interacts with the Google OR Tools library when the model is built using Builder.build/1 and when solved using Model.solve/1 or Model.solve/2 . References to the native objects are returned via NIF resources to the Elixir runtime as %Reference{} values. These are often stored in corresponding Exhort structs under the res key.","ref":"Exhort.html","title":"Exhort","type":"module"},{"doc":"Provide for the building of a model for eventual solving. All functions except build/0 are pure Elixir. Create a new builder with new/0 . Build a model with build/0 . build/0 interacts with the underlying native implementation, returning a Exhort.SAT.Model struct.","ref":"Exhort.SAT.Builder.html","title":"Exhort.SAT.Builder","type":"module"},{"doc":"Add an item or list of items to the builder.","ref":"Exhort.SAT.Builder.html#add/2","title":"Exhort.SAT.Builder.add/2","type":"function"},{"doc":"Build the model. Once the model is built it may be solved. This function interacts with the underlying native model.","ref":"Exhort.SAT.Builder.html#build/1","title":"Exhort.SAT.Builder.build/1","type":"function"},{"doc":"See Exhort.SAT.Constraint . Define a bounded constraint. The expression must include a boundary like == , &lt;= , &gt; , etc. x &lt; y The components of the expressoin may be simple mathematical expressions, including the use of + and * : x * y = z The sum/1 function may be used to sum over a series of terms: sum ( x + y ) == z The variables in the expression are defined in the model and do not by default reference the variables in Elixir scope. The pin operator, ^ may be used to reference a scoped Elixir variable. For example, where x is a model variable (e.g., def_int_var(x, {0, 3} )) and y is an Elixir variable (e.g., y = 2 ): x &lt; ^ y A for comprehension may be used to generate list values: sum ( for { x , y } &lt;- ^ list , do : x * y ) == z As a larger example: y = 20 z = [ { 0 , 1 } , { 2 , 3 } , { 4 , 5 } ] Builder . new ( ) |&gt; Builder . def_int_var ( x , { 0 , 3 } ) |&gt; Builder . constrain ( sum ( for { a , b } &lt;- ^ z , do : ^ a * ^ b ) &lt; y ) |&gt; Builder . build ( ) ...","ref":"Exhort.SAT.Builder.html#constrain/3","title":"Exhort.SAT.Builder.constrain/3","type":"macro"},{"doc":"Apply the constraint to the given list. See Exhort.SAT.Constraint for the list of constraints.","ref":"Exhort.SAT.Builder.html#constrain_list/4","title":"Exhort.SAT.Builder.constrain_list/4","type":"function"},{"doc":"Specifiy a decision strategy on a list of variables.","ref":"Exhort.SAT.Builder.html#decision_strategy/4","title":"Exhort.SAT.Builder.decision_strategy/4","type":"function"},{"doc":"Define a boolean variable in the model.","ref":"Exhort.SAT.Builder.html#def_bool_var/2","title":"Exhort.SAT.Builder.def_bool_var/2","type":"function"},{"doc":"Create a named constant. value should be a constant integer.","ref":"Exhort.SAT.Builder.html#def_constant/3","title":"Exhort.SAT.Builder.def_constant/3","type":"function"},{"doc":"Define an integer variable in the model. name is the variable name domain is the uppper and lower bounds of the integer as a tuple, {lower_bound, upper_bound}","ref":"Exhort.SAT.Builder.html#def_int_var/3","title":"Exhort.SAT.Builder.def_int_var/3","type":"function"},{"doc":"Define an interval variable in the model. See https://developers.google.com/optimization/reference/python/sat/python/cp_model#intervalvar name is the variable name start is the start of the interval size is the size of the interval stop is the end of the interval opts may specify if: bool_var , where bool_var being a previously defined boolean variable","ref":"Exhort.SAT.Builder.html#def_interval_var/6","title":"Exhort.SAT.Builder.def_interval_var/6","type":"function"},{"doc":"Add a constraint on the variable named by literal to the list of items in list .","ref":"Exhort.SAT.Builder.html#max_equality/3","title":"Exhort.SAT.Builder.max_equality/3","type":"function"},{"doc":"Specify an objective to maximize expression .","ref":"Exhort.SAT.Builder.html#maximize/2","title":"Exhort.SAT.Builder.maximize/2","type":"macro"},{"doc":"Specify an objective to minimize literal .","ref":"Exhort.SAT.Builder.html#minimize/2","title":"Exhort.SAT.Builder.minimize/2","type":"macro"},{"doc":"Start a new builder.","ref":"Exhort.SAT.Builder.html#new/0","title":"Exhort.SAT.Builder.new/0","type":"function"},{"doc":"","ref":"Exhort.SAT.Builder.html#t:t/0","title":"Exhort.SAT.Builder.t/0","type":"type"},{"doc":"A constraint on the model. Suggestion Consider using Exhort.SAT.Expr instead of using this module directly. The binary constraints are: :&lt; | :&lt;= | :== | :&gt;= | :&gt; | :&quot;abs==&quot; The list constraints are: :&quot;all!=&quot; | :no_overlap The expression must include a boundary: &lt; , &lt;= , == , &gt;= , &gt; . &quot;x&quot; &lt; &quot;y&quot; The components of the expressoin may be simple mathematical expressions, including the use of + and * : &quot;x&quot; * &quot;y&quot; = &quot;z&quot; The sum/1 function may be used to sum over a series of terms: sum ( &quot;x&quot; + &quot;y&quot; ) == &quot;z&quot; The variables in the expression may be model variables or Elixir variables. For example, where &quot;x&quot; is a model variable (e.g., def_int_var(x, {0, 3} )) and y is an Elixir variable (e.g., y = 2 ): &quot;x&quot; &lt; y A for comprehension may be used to generate list values: sum ( for { x , y } &lt;- list , do : &quot;x&quot; * &quot;y&quot; ) == &quot;z&quot; As a larger example: y = 20 z = [ { 0 , 1 } , { 2 , 3 } , { 4 , 5 } ] Builder . new ( ) |&gt; Builder . def_int_var ( &quot;x&quot; , { 0 , 3 } ) |&gt; Builder . constrain ( sum ( for { a , b } &lt;- z , do : a * b ) &lt; &quot;y&quot; ) |&gt; Builder . build ( ) ...","ref":"Exhort.SAT.Constraint.html","title":"Exhort.SAT.Constraint","type":"module"},{"doc":"Create a constraint that ensures each item in the list is different in the solution.","ref":"Exhort.SAT.Constraint.html#all_different/2","title":"Exhort.SAT.Constraint.all_different/2","type":"function"},{"doc":"Create longical and constraint on the list of booleans.","ref":"Exhort.SAT.Constraint.html#bool_and/1","title":"Exhort.SAT.Constraint.bool_and/1","type":"function"},{"doc":"Create a constraint that requires one of the booleans in the list to be true.","ref":"Exhort.SAT.Constraint.html#bool_or/1","title":"Exhort.SAT.Constraint.bool_or/1","type":"function"},{"doc":"Add an implication constraint where bool1 implies bool2 .","ref":"Exhort.SAT.Constraint.html#implication/2","title":"Exhort.SAT.Constraint.implication/2","type":"function"},{"doc":"Define a bounded constraint.","ref":"Exhort.SAT.Constraint.html#new/2","title":"Exhort.SAT.Constraint.new/2","type":"macro"},{"doc":"Create a constraint that ensures no overlap among the variables.","ref":"Exhort.SAT.Constraint.html#no_overlap/2","title":"Exhort.SAT.Constraint.no_overlap/2","type":"function"},{"doc":"","ref":"Exhort.SAT.Constraint.html#t:constraint/0","title":"Exhort.SAT.Constraint.constraint/0","type":"type"},{"doc":"","ref":"Exhort.SAT.Constraint.html#t:t/0","title":"Exhort.SAT.Constraint.t/0","type":"type"},{"doc":"Create an expression in the expression language. If the expression contains a comparison operator, it will be a constraint. Otherwise, it will be a linear expression. Use the new/2 macro to create a new expression. Use the remaining functions to create variables and list-based constraints.","ref":"Exhort.SAT.Expr.html","title":"Exhort.SAT.Expr","type":"module"},{"doc":"Create a constraint on the list ensuring that each variable in the list has a different value.","ref":"Exhort.SAT.Expr.html#all_different/2","title":"Exhort.SAT.Expr.all_different/2","type":"function"},{"doc":"Create logical AND constraint on the list of booleans.","ref":"Exhort.SAT.Expr.html#bool_and/1","title":"Exhort.SAT.Expr.bool_and/1","type":"macro"},{"doc":"Create a constraint that requires one of the booleans in the list to be true.","ref":"Exhort.SAT.Expr.html#bool_or/1","title":"Exhort.SAT.Expr.bool_or/1","type":"macro"},{"doc":"Define a new boolean variable. It must later be added to the model.","ref":"Exhort.SAT.Expr.html#def_bool_var/1","title":"Exhort.SAT.Expr.def_bool_var/1","type":"function"},{"doc":"Definite a constant. It must later be added to the model.","ref":"Exhort.SAT.Expr.html#def_constant/2","title":"Exhort.SAT.Expr.def_constant/2","type":"function"},{"doc":"Define a new integer variable. It must later be added to the model.","ref":"Exhort.SAT.Expr.html#def_int_var/2","title":"Exhort.SAT.Expr.def_int_var/2","type":"function"},{"doc":"Define an interval variable. It must later be added to the model. See Exhort.SAT.Builder.def_interval_var/6 . name is the variable name start is the start of the interval size is the size of the interval stop is the end of the interval","ref":"Exhort.SAT.Expr.html#def_interval_var/5","title":"Exhort.SAT.Expr.def_interval_var/5","type":"function"},{"doc":"Add an implication constraint where bool1 implies bool2 .","ref":"Exhort.SAT.Expr.html#implication/2","title":"Exhort.SAT.Expr.implication/2","type":"macro"},{"doc":"Create a new expression using the DSL. If the expression contains a comparison operator, it becomes a constraint.","ref":"Exhort.SAT.Expr.html#new/2","title":"Exhort.SAT.Expr.new/2","type":"macro"},{"doc":"Create a constraint on the list ensuring there are no overlap among the variables in the list.","ref":"Exhort.SAT.Expr.html#no_overlap/2","title":"Exhort.SAT.Expr.no_overlap/2","type":"function"},{"doc":"The model built from the Builder .","ref":"Exhort.SAT.Model.html","title":"Exhort.SAT.Model","type":"module"},{"doc":"Solve the model, returning the solution. This may only be called after the build function has been called.","ref":"Exhort.SAT.Model.html#solve/1","title":"Exhort.SAT.Model.solve/1","type":"function"},{"doc":"Solve the model, using a callback for each response to the model. The given function will be called on each improving feasible solution found during the search. For a non-optimization problem, if the option to find all solution was set, then this will be called on each new solution.","ref":"Exhort.SAT.Model.html#solve/2","title":"Exhort.SAT.Model.solve/2","type":"function"},{"doc":"","ref":"Exhort.SAT.Model.html#t:t/0","title":"Exhort.SAT.Model.t/0","type":"type"},{"doc":"A response from solving a model. Provides functions for retrieving variable values from the response. The set of valid variables are those defined in the model that was solved.","ref":"Exhort.SAT.SolverResponse.html","title":"Exhort.SAT.SolverResponse","type":"module"},{"doc":"Get the corresponding value of the boolean variable.","ref":"Exhort.SAT.SolverResponse.html#bool_val/2","title":"Exhort.SAT.SolverResponse.bool_val/2","type":"function"},{"doc":"","ref":"Exhort.SAT.SolverResponse.html#build/2","title":"Exhort.SAT.SolverResponse.build/2","type":"function"},{"doc":"Get the corresponding value of the integer variable.","ref":"Exhort.SAT.SolverResponse.html#int_val/2","title":"Exhort.SAT.SolverResponse.int_val/2","type":"function"},{"doc":"A map of the response metadata, :status , :objective , :walltime , :usertime .","ref":"Exhort.SAT.SolverResponse.html#stats/1","title":"Exhort.SAT.SolverResponse.stats/1","type":"function"},{"doc":"Get a variable value from the response.","ref":"Exhort.SAT.SolverResponse.html#value/2","title":"Exhort.SAT.SolverResponse.value/2","type":"function"},{"doc":"","ref":"Exhort.SAT.SolverResponse.html#t:t/0","title":"Exhort.SAT.SolverResponse.t/0","type":"type"},{"doc":"A boolean variable defined in the model.","ref":"Exhort.SAT.BoolVar.html","title":"Exhort.SAT.BoolVar","type":"module"},{"doc":"Define a new boolean variable. In the model, true and false are represented as 1 and 0, respectively. name - The variable name that may be referenced in other expressions.","ref":"Exhort.SAT.BoolVar.html#new/1","title":"Exhort.SAT.BoolVar.new/1","type":"function"},{"doc":"","ref":"Exhort.SAT.BoolVar.html#t:t/0","title":"Exhort.SAT.BoolVar.t/0","type":"type"},{"doc":"An integer variable defined in the model.","ref":"Exhort.SAT.IntVar.html","title":"Exhort.SAT.IntVar","type":"module"},{"doc":"Define a new integer variable. name - The variable name that may be referenced in other expressions. domain - The upper and lower bounds of the variable defined as a tuple, {lower_bound, upper_bound} .","ref":"Exhort.SAT.IntVar.html#new/2","title":"Exhort.SAT.IntVar.new/2","type":"function"},{"doc":"","ref":"Exhort.SAT.IntVar.html#t:t/0","title":"Exhort.SAT.IntVar.t/0","type":"type"},{"doc":"An interval variable defined in the model.","ref":"Exhort.SAT.IntervalVar.html","title":"Exhort.SAT.IntervalVar","type":"module"},{"doc":"Define a new interval variable. name - The variable name that may be referenced in other expressions. start - The lower bound of the interval size - The step size of the parts of the iterval stop - The upper bound of the interval opts - Supports if: bool as a condition on the interval constraint","ref":"Exhort.SAT.IntervalVar.html#new/5","title":"Exhort.SAT.IntervalVar.new/5","type":"function"},{"doc":"","ref":"Exhort.SAT.IntervalVar.html#t:t/0","title":"Exhort.SAT.IntervalVar.t/0","type":"type"},{"doc":"","ref":"binpacking.html","title":"Binpacking","type":"extras"},{"doc":"use Exhort.SAT.Builder","ref":"binpacking.html#setup","title":"Binpacking - Setup","type":"extras"},{"doc":"bin_capacity = 100 slack_capacity = 20 num_bins = 5 all_bins = Range . new ( 1 , num_bins ) items = [ { 20 , 6 } , { 15 , 6 } , { 30 , 4 } , { 45 , 3 } ]","ref":"binpacking.html#data","title":"Binpacking - Data","type":"extras"},{"doc":"xs = items |&gt; Enum . map ( fn { item , num_copies } -&gt; all_bins |&gt; Enum . map ( fn bin -&gt; IntVar . new ( &quot;x_ \#{ item } _ \#{ bin } &quot; , { 0 , num_copies } ) end ) end ) |&gt; List . flatten ( ) loads = all_bins |&gt; Enum . map ( fn bin -&gt; IntVar . new ( &quot;load_ \#{ bin } &quot; , { 0 , bin_capacity } ) end ) slacks = all_bins |&gt; Enum . map ( fn bin -&gt; BoolVar . new ( &quot;slack_ \#{ bin } &quot; ) end ) constrain_load_to_x = all_bins |&gt; Enum . map ( fn bin -&gt; expr = Enum . map ( items , &amp; { elem ( &amp;1 , 0 ) , &quot;x_ \#{ elem ( &amp;1 , 0 ) } _ \#{ bin } &quot; } ) load_bin = &quot;load_ \#{ bin } &quot; Constraint . new ( sum ( for { item , x } &lt;- expr , do : item * x ) == load_bin ) end ) placements = items |&gt; Enum . map ( fn { item , num_copies } -&gt; x_i = Enum . map ( all_bins , &amp; &quot;x_ \#{ item } _ \#{ &amp;1 } &quot; ) Constraint . new ( sum ( x_i ) == num_copies ) end ) constrain_load_to_slack = all_bins |&gt; Enum . map ( fn bin -&gt; safe_capacity = bin_capacity - slack_capacity load_bin = &quot;load_ \#{ bin } &quot; slack_bin = &quot;slack_ \#{ bin } &quot; [ Constraint . new ( load_bin &lt;= safe_capacity , if : slack_bin ) , Constraint . new ( load_bin &gt; safe_capacity , unless : slack_bin ) ] end ) |&gt; List . flatten ( )","ref":"binpacking.html#constraints","title":"Binpacking - Constraints","type":"extras"},{"doc":"builder = Builder . new ( ) |&gt; Builder . add ( xs ) |&gt; Builder . add ( loads ) |&gt; Builder . add ( slacks ) |&gt; Builder . add ( constrain_load_to_x ) |&gt; Builder . add ( placements ) |&gt; Builder . add ( constrain_load_to_slack ) |&gt; then ( fn builder -&gt; bins = Enum . map ( all_bins , &amp; &quot;slack_ \#{ &amp;1 } &quot; ) Builder . maximize ( builder , sum ( bins ) ) end ) response = builder |&gt; Builder . build ( ) |&gt; Model . solve ( )","ref":"binpacking.html#solve","title":"Binpacking - Solve","type":"extras"},{"doc":"","ref":"nurse-scheduling.html","title":"Nurse Scheduling","type":"extras"},{"doc":"Mix . install ( [ { :exhort , git : &quot;https://github.com/elixir-or-tools/exhort&quot; } , { :vega_lite , &quot;~&gt; 0.1.3&quot; } , { :kino , &quot;~&gt; 0.5.0&quot; } ] ) use Exhort.SAT.Builder alias VegaLite , as : Vl","ref":"nurse-scheduling.html#setup","title":"Nurse Scheduling - Setup","type":"extras"},{"doc":"num_nurses = 5 num_shifts = 3 num_days = 7 all_nurses = 1 .. num_nurses all_shifts = 1 .. num_shifts all_days = 1 .. num_days shift_requests = [ [ [ 0 , 0 , 1 ] , [ 0 , 0 , 0 ] , [ 0 , 0 , 0 ] , [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] , [ 0 , 1 , 0 ] , [ 0 , 0 , 1 ] ] , [ [ 0 , 0 , 0 ] , [ 0 , 0 , 0 ] , [ 0 , 1 , 0 ] , [ 0 , 1 , 0 ] , [ 1 , 0 , 0 ] , [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] ] , [ [ 0 , 1 , 0 ] , [ 0 , 1 , 0 ] , [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , [ 0 , 0 , 0 ] , [ 0 , 1 , 0 ] , [ 0 , 0 , 0 ] ] , [ [ 0 , 0 , 1 ] , [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , [ 0 , 1 , 0 ] , [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , [ 0 , 0 , 0 ] ] , [ [ 0 , 0 , 0 ] , [ 0 , 0 , 1 ] , [ 0 , 1 , 0 ] , [ 0 , 0 , 0 ] , [ 1 , 0 , 0 ] , [ 0 , 1 , 0 ] , [ 0 , 0 , 0 ] ] ] Shift variables: {n, d, s} : nurse n works shift s on day d . shifts = Enum . map ( all_nurses , fn nurse -&gt; Enum . map ( all_days , fn day -&gt; Enum . map ( all_shifts , fn shift -&gt; { nurse , day , shift } end ) end ) end ) |&gt; List . flatten ( ) shift_vars = shifts |&gt; Enum . map ( fn { nurse , day , shift } -&gt; Expr . def_bool_var ( &quot;shift_ \#{ nurse } _ \#{ day } _ \#{ shift } &quot; ) end ) Each shift is assigned to exactly one nurse in the schedule period. shift_nurses_per_period = all_days |&gt; Enum . map ( fn day -&gt; all_shifts |&gt; Enum . map ( fn shift -&gt; shift_options = Enum . filter ( shifts , fn { _n , d , s } -&gt; d == day and s == shift end ) shift_option_vars = Enum . map ( shift_options , fn { n , d , s } -&gt; &quot;shift_ \#{ n } _ \#{ d } _ \#{ s } &quot; end ) Expr . new ( sum ( shift_option_vars ) == 1 ) end ) end ) |&gt; List . flatten ( ) Each nurse works at most one shift per day nurse_shifts_per_day = all_nurses |&gt; Enum . map ( fn nurse -&gt; all_days |&gt; Enum . map ( fn day -&gt; shift_options = Enum . filter ( shifts , fn { n , d , _s } -&gt; n == nurse and d == day end ) shift_option_vars = Enum . map ( shift_options , fn { n , d , s } -&gt; &quot;shift_ \#{ n } _ \#{ d } _ \#{ s } &quot; end ) Expr . new ( sum ( shift_option_vars ) &lt;= 1 ) end ) end ) |&gt; List . flatten ( ) Try to distribute the shifts evenly, so that each nurse works min_shifts_per_nurse shifts. If this is not possible, because the total number of shifts is not divisible by the number of nurses, some nurses will be assigned one more shift. min_shifts_per_nurse = div ( num_shifts * num_days , num_nurses ) max_shifts_per_nurse = if rem ( num_shifts * num_days , num_nurses ) == 0 do min_shifts_per_nurse else min_shifts_per_nurse + 1 end distribution_constraints = all_nurses |&gt; Enum . map ( fn nurse -&gt; shift_options = Enum . filter ( shifts , fn { n , _d , _s } -&gt; n == nurse end ) shift_option_vars = Enum . map ( shift_options , fn { n , d , s } -&gt; &quot;shift_ \#{ n } _ \#{ d } _ \#{ s } &quot; end ) [ Expr . new ( min_shifts_per_nurse &lt;= sum ( shift_option_vars ) ) , Expr . new ( sum ( shift_option_vars ) &lt;= max_shifts_per_nurse ) ] end ) |&gt; List . flatten ( )","ref":"nurse-scheduling.html#data","title":"Nurse Scheduling - Data","type":"extras"},{"doc":"max_list = all_nurses |&gt; Enum . map ( fn nurse -&gt; all_days |&gt; Enum . map ( fn day -&gt; all_shifts |&gt; Enum . map ( fn shift -&gt; shift_var = &quot;shift_ \#{ nurse } _ \#{ day } _ \#{ shift } &quot; request = shift_requests |&gt; Enum . at ( nurse - 1 ) |&gt; Enum . at ( day - 1 ) |&gt; Enum . at ( shift - 1 ) Expr . new ( shift_var * request ) end ) end ) end ) |&gt; List . flatten ( ) response = Builder . new ( ) |&gt; Builder . add ( shift_vars ) |&gt; Builder . add ( shift_nurses_per_period ) |&gt; Builder . add ( nurse_shifts_per_day ) |&gt; Builder . add ( distribution_constraints ) |&gt; Builder . maximize ( sum ( max_list ) ) |&gt; Builder . build ( ) |&gt; Model . solve ( ) Optimal result is: [ 5 , 4 , 4 , 4 , 4 ] shift_counts = Enum . reduce ( all_nurses , [ ] , fn nurse , acc -&gt; count = Enum . reduce ( all_days , 0 , fn day , acc -&gt; Enum . reduce ( all_shifts , acc , fn shift , acc -&gt; if SolverResponse . bool_val ( response , &quot;shift_ \#{ nurse } _ \#{ day } _ \#{ shift } &quot; ) do acc + 1 else acc end end ) end ) acc ++ [ count ] end ) values = Enum . map ( all_nurses , fn nurse -&gt; Enum . map ( all_days , fn day -&gt; Enum . map ( all_shifts , fn shift -&gt; nurse_works? = SolverResponse . bool_val ( response , &quot;shift_ \#{ nurse } _ \#{ day } _ \#{ shift } &quot; ) nurse = if nurse_works? , do : nurse , else : nil %{ day : day , shift : shift , nurse : nurse } end ) end ) end ) |&gt; List . flatten ( ) Vl . new ( width : 400 , height : 400 ) |&gt; Vl . data_from_values ( values ) |&gt; Vl . mark ( :rect , tooltip : true ) |&gt; Vl . encode_field ( :x , &quot;shift&quot; , type : :ordinal ) |&gt; Vl . encode_field ( :y , &quot;day&quot; , type : :ordinal ) |&gt; Vl . encode_field ( :color , &quot;nurse&quot; , type : :quantitative )","ref":"nurse-scheduling.html#objective-function","title":"Nurse Scheduling - Objective Function","type":"extras"},{"doc":"","ref":"channeling.html","title":"Binpacking","type":"extras"},{"doc":"use Exhort.SAT.Builder","ref":"channeling.html#setup","title":"Binpacking - Setup","type":"extras"},{"doc":"builder = Builder . new ( ) |&gt; Builder . def_int_var ( :x , { 0 , 10 } ) |&gt; Builder . def_int_var ( :y , { 0 , 10 } ) |&gt; Builder . def_bool_var ( :b ) |&gt; Builder . constrain ( :x &gt;= 5 , if : :b ) |&gt; Builder . constrain ( :x &lt; 5 , unless : :b ) |&gt; Builder . constrain ( :x + :y == 10 , if : :b ) |&gt; Builder . constrain ( :y == 0 , unless : :b )","ref":"channeling.html#model","title":"Binpacking - Model","type":"extras"},{"doc":"{ response , acc } = builder |&gt; Builder . build ( ) |&gt; Model . solve ( fn _response , nil -&gt; 1 _response , acc -&gt; acc + 1 end )","ref":"channeling.html#solution","title":"Binpacking - Solution","type":"extras"},{"doc":"","ref":"rabbits-and-pheasants.html","title":"Rabbits and Pheasants","type":"extras"},{"doc":"use Exhort.SAT.Builder","ref":"rabbits-and-pheasants.html#setup","title":"Rabbits and Pheasants - Setup","type":"extras"},{"doc":"response = Builder . new ( ) |&gt; Builder . def_int_var ( &quot;r&quot; , { 0 , 100 } ) |&gt; Builder . def_int_var ( &quot;p&quot; , { 0 , 100 } ) |&gt; Builder . constrain ( &quot;r&quot; + &quot;p&quot; == 20 ) |&gt; Builder . constrain ( 4 * &quot;r&quot; + 2 * &quot;p&quot; == 56 ) |&gt; Builder . build ( ) |&gt; Model . solve ( )","ref":"rabbits-and-pheasants.html#model","title":"Rabbits and Pheasants - Model","type":"extras"},{"doc":"","ref":"minimal-job-shop.html","title":"Minimal Job Shop","type":"extras"},{"doc":"use Exhort.SAT.Builder","ref":"minimal-job-shop.html#setup","title":"Minimal Job Shop - Setup","type":"extras"},{"doc":"# task = (machine_id, processing_time). jobs_data = [ # Job0 [ { 0 , 3 } , { 1 , 2 } , { 2 , 2 } ] , # Job1 [ { 0 , 2 } , { 2 , 1 } , { 1 , 4 } ] , # Job2 [ { 1 , 4 } , { 2 , 3 } ] ] # Computes horizon dynamically as the sum of all durations. horizon = jobs_data |&gt; List . flatten ( ) |&gt; Enum . map ( &amp; elem ( &amp;1 , 1 ) ) |&gt; Enum . sum ( ) # [ # {0, [{0, {0, 3}}, {1, {1, 2}}, {2, {2, 2}}]}, # {1, [{0, {0, 2}}, {1, {2, 1}}, {2, {1, 4}}]}, # {2, [{0, {1, 4}}, {1, {2, 3}}]} # ] jobs = jobs_data |&gt; Enum . with_index ( ) |&gt; Enum . map ( fn { job , job_id } -&gt; { job_id , job |&gt; Enum . with_index ( ) |&gt; Enum . map ( fn { { machine_id , processing_time } , task_id } -&gt; { task_id , { machine_id , processing_time } } end ) } end ) # %{{job_id, task_id} =&gt; %{machine: _, start: _, processing_time: _, end: _, interval: _}} tasks = jobs |&gt; Enum . map ( fn { job_id , job } -&gt; job |&gt; Enum . map ( fn { task_id , { machine_id , processing_time } } -&gt; suffix = &quot; \#{ job_id } _ \#{ task_id } &quot; { { job_id , task_id } , %{ machine : machine_id , start : &quot;start_ \#{ suffix } &quot; , processing_time : processing_time , end : &quot;end_ \#{ suffix } &quot; , interval : &quot;interval_ \#{ suffix } &quot; } } end ) end ) |&gt; List . flatten ( ) |&gt; Enum . into ( %{ } )","ref":"minimal-job-shop.html#data","title":"Minimal Job Shop - Data","type":"extras"},{"doc":"machine_constraints = tasks |&gt; Enum . map ( fn { _ , %{ machine : machine_id , interval : interval_var } } -&gt; { machine_id , interval_var } end ) |&gt; Enum . group_by ( fn { machine_id , _ } -&gt; machine_id end , fn { _ , interval_var } -&gt; interval_var end ) |&gt; Enum . map ( fn { _machine_id , intervals } -&gt; Constraint . no_overlap ( intervals ) end ) task_vars = tasks |&gt; Map . values ( ) |&gt; Enum . map ( fn %{ start : start_var , processing_time : processing_time , end : end_var , interval : interval_var } -&gt; [ IntVar . new ( start_var , { 0 , horizon } ) , IntVar . new ( end_var , { 0 , horizon } ) , IntervalVar . new ( interval_var , start_var , processing_time , end_var ) ] end ) |&gt; List . flatten ( ) task_constraints = jobs |&gt; Enum . map ( fn { job_id , job } -&gt; job |&gt; Enum . slice ( 0 , length ( job ) - 1 ) |&gt; Enum . map ( fn { task_id , _task } -&gt; task_start = tasks [ { job_id , task_id + 1 } ] . start task_end = tasks [ { job_id , task_id } ] . end Constraint . new ( task_start &gt;= task_end ) end ) end ) |&gt; List . flatten ( )","ref":"minimal-job-shop.html#constraints","title":"Minimal Job Shop - Constraints","type":"extras"},{"doc":"builder = Builder . new ( ) |&gt; Builder . add ( task_vars ) |&gt; Builder . add ( machine_constraints ) |&gt; Builder . add ( task_constraints ) |&gt; Builder . def_int_var ( &quot;makespan&quot; , { 0 , horizon } ) |&gt; Builder . max_equality ( &quot;makespan&quot; , jobs |&gt; Enum . map ( fn { job_id , job } -&gt; tasks [ { job_id , length ( job ) - 1 } ] . end end ) ) |&gt; Builder . minimize ( &quot;makespan&quot; ) response = builder |&gt; Builder . build ( ) |&gt; Model . solve ( ) tasks |&gt; Enum . map ( fn { { job_id , task_id } , %{ machine : machine_id , start : start_var , end : end_var } } -&gt; { &quot;Machine: \#{ machine_id } &quot; , { &quot; \#{ job_id } _ \#{ task_id } &quot; , SolverResponse . int_val ( response , start_var ) , SolverResponse . int_val ( response , end_var ) } } end ) |&gt; Enum . group_by ( fn { machine , _ } -&gt; machine end ) |&gt; Enum . map ( fn { machine , jobs } -&gt; { machine , jobs |&gt; Enum . map ( fn { _machine , { id , task_start , task_end } } -&gt; { id , { task_start , task_end } } end ) |&gt; Enum . into ( %{ } ) } end ) |&gt; Enum . into ( %{ } )","ref":"minimal-job-shop.html#solve","title":"Minimal Job Shop - Solve","type":"extras"},{"doc":"","ref":"ranking-sample-sat.html","title":"Ranking Sample","type":"extras"},{"doc":"Code sample to demonstrates how to rank intervals. use Exhort.SAT.Builder","ref":"ranking-sample-sat.html#section","title":"Ranking Sample - Section","type":"extras"},{"doc":"Ranks tasks in a no_overlap constraint. horizon = 100 num_tasks = 4 all_tasks = 0 .. 3 %{ starts : starts , ends : ends , presences : presences , presence_expr : presences_expr , o_intervals : o_intervals , ranks : ranks } = for task &lt;- all_tasks do duration = task + 1 presence = task &lt; div ( num_tasks , 2 ) %{ starts : { task , Expr . def_int_var ( &quot;starts_ \#{ task } &quot; , { 0 , horizon } ) } , ends : { task , Expr . def_int_var ( &quot;ends_ \#{ task } &quot; , { 0 , horizon } ) } , presences : { task , Expr . def_bool_var ( &quot;presence_ \#{ task } &quot; ) } , presence_expr : { task , if ( presence , do : Expr . new ( &quot;presence_ \#{ task } &quot; == 1 ) ) } , o_intervals : { task , Expr . def_interval_var ( &quot;o_interval_ \#{ task } &quot; , &quot;starts_ \#{ task } &quot; , duration , &quot;ends_ \#{ task } &quot; , if : &quot;presence_ \#{ task } &quot; ) } , ranks : { task , Expr . def_int_var ( &quot;ranks_ \#{ task } &quot; , { - 1 , num_tasks - 1 } ) } } end |&gt; Enum . reduce ( %{ } , fn task_vars , merged -&gt; Map . merge ( merged , task_vars , fn _ , merged_var , { _ , nil } -&gt; merged_var _ , merged_var , new_var when is_list ( merged_var ) -&gt; [ new_var | merged_var ] _ , merged_var , new_var -&gt; [ new_var , merged_var ] end ) end ) |&gt; Enum . map ( fn { key , task_vars } -&gt; { key , Enum . into ( task_vars , %{ } ) } end ) |&gt; Enum . into ( %{ } ) no_overlap_intervals = Expr . no_overlap ( for task &lt;- all_tasks , do : o_intervals [ task ] )","ref":"ranking-sample-sat.html#ranking-sample-sat","title":"Ranking Sample - Ranking Sample Sat","type":"extras"},{"doc":"Add constraints and variables to links tasks and ranks. Assumes that all starts are disjoint, meaning that all tasks have a strictly positive duration, and they appear in the same :no_overlap constraint. Precedence variables Creates precedence variables between pairs of intervals. presedence_defs = for i &lt;- all_tasks , j &lt;- all_tasks do if i == j do { { i , j } , presences [ i ] } else i_before_j = &quot; \#{ i } before \#{ j } &quot; { Expr . def_bool_var ( i_before_j ) , Expr . new ( starts [ i ] &lt; starts [ j ] , if : i_before_j ) , { { i , j } , i_before_j } } end end precedences = presedence_defs |&gt; Enum . map ( fn { _ , _ , ij } -&gt; ij ij -&gt; ij end ) |&gt; Enum . into ( %{ } ) precedence_constraints = presedence_defs |&gt; Enum . filter ( fn { _ , _ } -&gt; false { _ , _ , _ } -&gt; true end ) |&gt; Enum . map ( fn { bool , const , _ } -&gt; [ bool , const ] end ) |&gt; List . flatten ( ) Optional intervals optional_intervals = for i &lt;- Enum . take ( all_tasks , Range . size ( all_tasks ) - 1 ) , j &lt;- Range . new ( i + 1 , num_tasks - 1 ) do [ Expr . implication ( not presences [ i ] , not precedences [ { i , j } ] ) , Expr . implication ( not presences [ i ] , not precedences [ { j , i } ] ) , Expr . implication ( not presences [ j ] , not precedences [ { i , j } ] ) , Expr . bool_or ( [ precedences [ { i , j } ] , precedences [ { j , i } ] , not presences [ i ] , not presences [ j ] ] ) , Expr . implication ( precedences [ { i , j } ] , not precedences [ { j , i } ] ) , Expr . implication ( precedences [ { j , i } ] , not precedences [ { i , j } ] ) ] end |&gt; List . flatten ( ) Links precedences and ranks precedences_to_ranks = for i &lt;- all_tasks do Expr . new ( ranks [ i ] == sum ( for j &lt;- all_tasks , do : precedences [ { j , i } ] ) - 1 ) end rank_0 = Expr . new ( ranks [ 0 ] &lt; ranks [ 1 ] ) makespan = Expr . def_int_var ( &quot;makespan&quot; , { 0 , horizon } ) makespan_constraints = for task &lt;- all_tasks do Expr . new ( ends [ task ] &lt;= makespan , if : presences [ task ] ) end","ref":"ranking-sample-sat.html#rank-tasks","title":"Ranking Sample - Rank Tasks","type":"extras"},{"doc":"Solve the model. response = Builder . new ( ) |&gt; Builder . add ( Map . values ( starts ) ) |&gt; Builder . add ( Map . values ( ends ) ) |&gt; Builder . add ( Map . values ( presences ) ) |&gt; Builder . add ( Map . values ( presences_expr ) ) |&gt; Builder . add ( Map . values ( o_intervals ) ) |&gt; Builder . add ( Map . values ( ranks ) ) |&gt; Builder . add ( no_overlap_intervals ) |&gt; Builder . add ( precedence_constraints ) |&gt; Builder . add ( optional_intervals ) |&gt; Builder . add ( precedences_to_ranks ) |&gt; Builder . add ( rank_0 ) |&gt; Builder . add ( makespan ) |&gt; Builder . add ( makespan_constraints ) |&gt; Builder . minimize ( 2 * makespan - 7 * sum ( for task &lt;- all_tasks , do : presences [ task ] ) ) |&gt; Builder . build ( ) |&gt; Model . solve ( ) SolverResponse . stats ( response ) IO . puts ( &quot;makespan: \#{ SolverResponse . int_val ( response , &quot;makespan&quot; ) } &quot; ) for task &lt;- all_tasks do if SolverResponse . bool_val ( response , presences [ task ] ) do IO . puts ( &quot;task \#{ task } starts at \#{ SolverResponse . int_val ( response , starts [ task ] ) } and has rank \#{ SolverResponse . int_val ( response , ranks [ task ] ) } &quot; ) else IO . puts ( &quot;task \#{ task } is not performed and has rank \#{ SolverResponse . int_val ( response , ranks [ task ] ) } &quot; ) end end","ref":"ranking-sample-sat.html#solution","title":"Ranking Sample - Solution","type":"extras"},{"doc":"","ref":"multiple-knapsack.html","title":"Multiple Knapsack","type":"extras"},{"doc":"Mix . install ( [ { :exhort , git : &quot;https://github.com/elixir-or-tools/exhort&quot; } , { :vega_lite , &quot;~&gt; 0.1.3&quot; } , { :kino , &quot;~&gt; 0.5.0&quot; } ] ) use Exhort.SAT.Builder alias VegaLite , as : Vl","ref":"multiple-knapsack.html#setup","title":"Multiple Knapsack - Setup","type":"extras"},{"doc":"weights = [ 48 , 30 , 42 , 36 , 36 , 48 , 42 , 42 , 36 , 24 , 30 , 30 , 42 , 36 , 36 ] values = [ 10 , 30 , 25 , 50 , 35 , 30 , 15 , 40 , 30 , 35 , 45 , 10 , 20 , 30 , 25 ] num_items = length ( weights ) bin_capacities = [ 100 , 100 , 100 , 100 , 100 ] num_bins = length ( bin_capacities ) all_items = Range . new ( 0 , num_items - 1 ) all_bins = Range . new ( 0 , num_bins - 1 ) { all_items , all_bins } x = for idx &lt;- all_items , b &lt;- all_bins do { { idx , b } , Expr . def_int_var ( &quot;x_ \#{ idx } _ \#{ b } &quot; , { 0 , 1 } ) } end |&gt; Enum . into ( %{ } ) max_value = Enum . sum ( values ) value[b] is the value of bin b when packed. value = for b &lt;- all_bins do { b , Expr . def_int_var ( &quot;value_ \#{ b } &quot; , { 0 , max_value } ) } end |&gt; Enum . into ( %{ } ) value_constraints = for b &lt;- all_bins do Expr . new ( value [ b ] == sum ( for i &lt;- all_items , do : x [ { i , b } ] * Enum . at ( values , i ) ) ) end","ref":"multiple-knapsack.html#model","title":"Multiple Knapsack - Model","type":"extras"},{"doc":"Each item can be in at most one bin. one_bin_per_item = for idx &lt;- all_items do Expr . new ( sum ( for b &lt;- all_bins , do : x [ { idx , b } ] ) &lt;= 1 ) end The amount packed in each bin cannot exceed its capacity. max_capacity_per_item = for b &lt;- all_bins do Expr . new ( sum ( for i &lt;- all_items , do : x [ { i , b } ] * Enum . at ( weights , i ) ) &lt;= Enum . at ( bin_capacities , b ) ) end builder = Builder . new ( ) |&gt; Builder . add ( Map . values ( x ) ) |&gt; Builder . add ( Map . values ( value ) ) |&gt; Builder . add ( value_constraints ) |&gt; Builder . add ( one_bin_per_item ) |&gt; Builder . add ( max_capacity_per_item ) |&gt; Builder . maximize ( sum ( Map . values ( value ) ) ) response = builder |&gt; Builder . build ( ) |&gt; Model . solve ( ) response |&gt; SolverResponse . stats ( )","ref":"multiple-knapsack.html#constraints","title":"Multiple Knapsack - Constraints","type":"extras"},{"doc":"Create a table-like data structure for VegaLite bin_items = for b &lt;- all_bins , idx &lt;- all_items do { b , idx } end |&gt; Enum . filter ( fn { b , idx } -&gt; SolverResponse . value ( response , x [ { idx , b } ] ) &gt; 0 end ) |&gt; Enum . map ( fn { b , idx } -&gt; %{ &quot;bin&quot; =&gt; b , &quot;item&quot; =&gt; idx , &quot;weight&quot; =&gt; Enum . at ( weights , idx ) , &quot;value&quot; =&gt; Enum . at ( values , idx ) } end ) Vl . new ( width : 650 , height : 500 ) |&gt; Vl . data_from_values ( bin_items ) |&gt; Vl . encode_field ( :x , &quot;bin&quot; , type : :nominal ) |&gt; Vl . encode_field ( :y , &quot;weight&quot; , aggregate : :sum , stack : &quot;zero&quot; ) |&gt; Vl . layers ( [ Vl . new ( ) |&gt; Vl . mark ( :bar ) |&gt; Vl . encode ( :color , field : &quot;item&quot; ) , Vl . new ( ) |&gt; Vl . mark ( :text , baseline : &quot;bottom&quot; , fill : &quot;black&quot; ) |&gt; Vl . encode_field ( :text , &quot;weight&quot; , type : :quantitative , aggregate : :sum ) |&gt; Vl . encode ( :color , field : &quot;item&quot; ) ] )","ref":"multiple-knapsack.html#visualize","title":"Multiple Knapsack - Visualize","type":"extras"}]